#!/bin/sh
./make_formation.py bg1 > ../gemrb/override/bg1/formatio.2da
./make_formation.py bg2 > ../gemrb/override/bg2/formatio.2da
./make_formation.py iwd > ../gemrb/override/iwd/formatio.2da
./make_formation.py iwd2 > ../gemrb/override/iwd2/formatio.2da
./make_formation.py pst > ../gemrb/override/pst/formatio.2da
./make_formation.py how > ../gemrb/override/how/formatio.2da

