/* GemRB - Infinity Engine Emulator
 * Copyright (C) 2003 The GemRB Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 *
 */

namespace GemRB {

#define FEAT_AEGIS_OF_RIME		0
#define FEAT_AMBIDEXTERITY		1
#define FEAT_AQUA_MORTIS		2
#define FEAT_ARMOUR_PROFICIENCY		3
#define FEAT_ARMORED_ARCANA		4
#define FEAT_ARTERIAL_STRIKE		5
#define FEAT_BLIND_FIGHT		6
#define FEAT_BULLHEADED			7
#define FEAT_CLEAVE			8
#define FEAT_COMBAT_CASTING		9
#define FEAT_COURTEOUS_MAGOCRACY	10
#define FEAT_CRIPPLING_STRIKE		11
#define FEAT_DASH			12
#define FEAT_DEFLECT_ARROWS		13
#define FEAT_DIRTY_FIGHTING		14
#define FEAT_DISCIPLINE			15
#define FEAT_DODGE			16
#define FEAT_ENVENOM_WEAPON		17
#define FEAT_EXOTIC_BASTARD		18
#define FEAT_EXPERTISE			19
#define FEAT_EXTRA_RAGE			20
#define FEAT_EXTRA_SHAPESHIFTING	21
#define FEAT_EXTRA_SMITING		22
#define FEAT_EXTRA_TURNING		23
#define FEAT_FIENDSLAYER		24
#define FEAT_FORESTER			25
#define FEAT_GREAT_FORTITUDE		26
#define FEAT_HAMSTRING			27
#define FEAT_HERETICS_BANE		28
#define FEAT_HEROIC_INSPIRATION		29
#define FEAT_IMPROVED_CRITICAL		30
#define FEAT_IMPROVED_EVASION		31
#define FEAT_IMPROVED_INITIATIVE	32
#define FEAT_IMPROVED_TURNING		33
#define FEAT_IRON_WILL			34
#define FEAT_LIGHTNING_REFLEXES		35
#define FEAT_LINGERING_SONG		36
#define FEAT_LUCK_OF_HEROES		37
#define FEAT_MARTIAL_AXE		38
#define FEAT_MARTIAL_BOW		39
#define FEAT_MARTIAL_FLAIL		40
#define FEAT_MARTIAL_GREATSWORD		41
#define FEAT_MARTIAL_HAMMER		42
#define FEAT_MARTIAL_LARGESWORD		43
#define FEAT_MARTIAL_POLEARM		44
#define FEAT_MAXIMIZED_ATTACKS		45
#define FEAT_MERCANTILE_BACKGROUND	46
#define FEAT_POWER_ATTACK		47
#define FEAT_PRECISE_SHOT		48
#define FEAT_RAPID_SHOT			49
#define FEAT_RESIST_POISON		50
#define FEAT_SCION_OF_STORMS		51
#define FEAT_SHIELD_PROF		52
#define FEAT_SIMPLE_CROSSBOW		53
#define FEAT_SIMPLE_MACE		54
#define FEAT_SIMPLE_MISSILE		55
#define FEAT_SIMPLE_QUARTERSTAFF	56
#define FEAT_SIMPLE_SMALLBLADE		57
#define FEAT_SLIPPERY_MIND		58
#define FEAT_SNAKE_BLOOD		59
#define FEAT_SPELL_FOCUS_ENCHANTMENT	60
#define FEAT_SPELL_FOCUS_EVOCATION	61
#define FEAT_SPELL_FOCUS_NECROMANCY	62
#define FEAT_SPELL_FOCUS_TRANSMUTE	63
#define FEAT_SPELL_PENETRATION		64
#define FEAT_SPIRIT_OF_FLAME		65
#define FEAT_STRONG_BACK		66
#define FEAT_STUNNING_FIST		67
#define FEAT_SUBVOCAL_CASTING		68
#define FEAT_TOUGHNESS			69
#define FEAT_TWO_WEAPON_FIGHTING	70
#define FEAT_WEAPON_FINESSE		71
#define FEAT_WILDSHAPE_BOAR		72
#define FEAT_WILDSHAPE_PANTHER		73
#define FEAT_WILDSHAPE_SHAMBLER		74

}
