/* GemRB - Infinity Engine Emulator
 * Copyright (C) 2011 The GemRB Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "Win32Console.h"

#include <cstdio>

#define ADV_TEXT
#include <conio.h>

namespace GemRB {

Win32ConsoleLogger::Win32ConsoleLogger(log_level level, bool useColor)
: StdioLogWriter(level, useColor)
{
	hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
	
	GetConsoleMode(hConsole, &dwMode);
	SetConsoleMode(hConsole, dwMode | ENABLE_VIRTUAL_TERMINAL_PROCESSING);
}

Win32ConsoleLogger::~Win32ConsoleLogger()
{
	SetConsoleMode(hConsole, dwMode);
}

Logger::WriterPtr createWin32ConsoleLogger()
{
	return Logger::WriterPtr(new Win32ConsoleLogger(DEBUG, true));
}

}
